/*-
 * Copyright (c) 2010 Ed Schouten <ed@80386.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <ncurses.h>
#include <stdlib.h>
#include <time.h>

#ifndef __unused
#define	__unused
#endif

struct card {
	unsigned char	number;
#define	NUM_LOWEST	0
#define	NUM_HIGHEST	12
	unsigned char	suit;
#define	SUIT_CLUBS	0
#define	SUIT_DIAMONDS	1
#define	SUIT_HEARTS	2
#define	SUIT_SPADES	3
#define	SUIT_SAME(a, b) \
	(((a) == SUIT_DIAMONDS || (a) == SUIT_HEARTS) == \
	((b) == SUIT_DIAMONDS || (b) == SUIT_HEARTS))
	char		visible;

	struct card	*next;
	struct card	*prev;
};

struct cardstack {
	struct card	*first;
	struct card	*last;
	int		type;
#define	CST_STOCK	0
#define	CST_WASTE	1
#define	CST_FOUNDATION	2
#define	CST_TABLEAU	3
};

static struct cardstack stock;
static struct cardstack waste;
#define	NFOUNDATIONS	4
static struct cardstack foundations[NFOUNDATIONS];
#define	NTABLEAUS	6
static struct cardstack tableaus[NTABLEAUS];
#define	NCARDS		52
static struct card cards[NCARDS];
static struct cardstack *stepfrom = NULL;

static void
cs_init(struct cardstack *cs, int type)
{

	cs->first = cs->last = NULL;
	cs->type = type;
}

static void
cs_insert_tail(struct cardstack *ts, struct card *cf)
{

	if (ts->last == NULL)
		ts->first = cf;
	else
		ts->last->next = cf;
	cf->prev = ts->last;
	cf->next = NULL;
	ts->last = cf;
}

static void
cs_move_tail(struct cardstack *fs, struct cardstack *ts, struct card *cf)
{
	struct card *cl = fs->last;

	/* Disconnect this chain from the old stack. */
	if (cf->prev == NULL)
		fs->first = NULL;
	else
		cf->prev->next = NULL;
	fs->last = cf->prev;

	/* Hook it up to the new stack. */
	if (ts->last == NULL)
		ts->first = cf;
	else
		ts->last->next = cf;
	cf->prev = ts->last;
	ts->last = cl;
}

static int
okay_foundation(const struct card *fl, const struct card *tl)
{

	/* Same suit, increasing order. */
	return ((tl == NULL && fl->number == NUM_LOWEST) ||
	    (tl != NULL && tl->suit == fl->suit &&
	    tl->number == fl->number - 1));
}

static int
okay_tableau(const struct card *fl, const struct card *tl)
{

	/* Alternating color, decreasing order. */
	return ((tl == NULL && fl->number == NUM_HIGHEST) ||
	    (tl != NULL && !SUIT_SAME(tl->suit, fl->suit) &&
	    tl->number == fl->number + 1));
}

static void
initstacks(void)
{
	int i;

	cs_init(&stock, CST_STOCK);
	cs_init(&waste, CST_WASTE);
	for (i = 0; i < NFOUNDATIONS; i++)
		cs_init(&foundations[i], CST_FOUNDATION);
	for (i = 0; i < NTABLEAUS; i++)
		cs_init(&tableaus[i], CST_TABLEAU);
}

static void
initcards(void)
{
	struct card t;
	int i, j;

	/* Default configuration. */
	for (i = 0; i < NCARDS; i++) {
		cards[i].number = i % (NUM_HIGHEST + 1);
		cards[i].suit = i / (NUM_HIGHEST + 1);
		cards[i].visible = 0;
	}

	/* Shuffle the cards. */
#if defined(__FreeBSD__) || defined(__APPLE__)
	srandomdev();
#else
	srandom(time(NULL));
#endif
	for (i = NCARDS - 1; i > 0; i--) {
		j = random() % i;
		t = cards[i];
		cards[i] = cards[j];
		cards[j] = t;
	}
}

static void
loadstacks(void)
{
	int i, j, c;

	/* Default layout. */
	c = 0;
	for (i = 0; i < NTABLEAUS; i++) {
		for (j = 0; j <= i; j++)
			cs_insert_tail(&tableaus[i], &cards[c++]);
		tableaus[i].last->visible = 1;
	}

	/* Store remainder in the stock. */
	for (; c < NCARDS; c++)
		cs_insert_tail(&stock, &cards[c]);
}

static void
initterm(void)
{

	initscr();
	nonl();
	cbreak();
	noecho();
	curs_set(0);

	start_color();
	init_pair(1, COLOR_RED, COLOR_WHITE);
	init_pair(2, COLOR_BLACK, COLOR_WHITE);
	init_pair(3, COLOR_WHITE, COLOR_BLUE);
}

static void
endterm(void)
{

	endwin();
}

static struct cardstack *
getstack(char c)
{

	switch (c) {
	case 'w':
		return (&waste);
	case 'e':
		return (&foundations[0]);
	case 'r':
		return (&foundations[1]);
	case 't':
		return (&foundations[2]);
	case 'y':
		return (&foundations[3]);
	case 'a':
		return (&tableaus[0]);
	case 's':
		return (&tableaus[1]);
	case 'd':
		return (&tableaus[2]);
	case 'f':
		return (&tableaus[3]);
	case 'g':
		return (&tableaus[4]);
	case 'h':
		return (&tableaus[5]);
	default:
		return (NULL);
	}
}

static void
displaystack(const struct cardstack *head, int top, int left)
{
	char const * const nmap[] = { "A", "2", "3", "4", "5",
	    "6", "7", "8", "9", "10", "J", "Q", "K" };
	struct card *c;
	char buf[6];
	int attrs, i;

	attrs = (stepfrom == head) ? A_BOLD : 0;
	for (c = head->first; c != NULL; c = c->next) {
		if (c->visible) {
			if (SUIT_SAME(c->suit, SUIT_HEARTS))
				attrset(COLOR_PAIR(1) | attrs);
			else
				attrset(COLOR_PAIR(2) | attrs);
		} else {
			attrset(COLOR_PAIR(3) | attrs);
		}
		mvaddch(top, left, ACS_ULCORNER);
		for (i = 0; i < 5; i++)
			addch(ACS_HLINE);
		addch(ACS_URCORNER);
		mvaddch(top + 1, left, ACS_VLINE);
		if (c->visible) {
			sprintf(buf, " %2s%c ", nmap[c->number],
			    "CDHS"[c->suit]);
			addstr(buf);
		} else {
			addstr(" --- ");
		}
		addch(ACS_VLINE);
		mvaddch(top + 2, left, ACS_LLCORNER);
		for (i = 0; i < 5; i++)
			addch(ACS_HLINE);
		addch(ACS_LRCORNER);

		/* Unfold the tableaus. */
		if (head->type == CST_TABLEAU)
			top += 2;
	}
}

static void
display(void)
{
	int i;

	erase();
	attrset(COLOR_PAIR(0));
	for (i = 0; i < 4; i++)
		mvaddch(i, 16, ACS_VLINE);
	for (i = 0; i < COLS; i++)
		mvaddch(4, i, i == 16 ? ACS_BTEE : ACS_HLINE);
	for (i = 0; i < 6; i++) {
		mvaddch(0, 3 + i * 9, "qwerty"[i]);
		mvaddch(5, 3 + i * 9, "asdfgh"[i]);
	}
	displaystack(&stock, 1, 0);
	displaystack(&waste, 1, 9);
	for (i = 0; i < NFOUNDATIONS; i++)
		displaystack(&foundations[i], 1, 9*2 + i * 9);
	for (i = 0; i < NTABLEAUS; i++)
		displaystack(&tableaus[i], 6, i * 9);
}

static int
finished(void)
{
	struct card *c;
	int i;

	/* Terminate when all cards on the tableaus are visible. */
	for (i = 0; i < NTABLEAUS; i++) {
		c = tableaus[i].first;
		if (c != NULL && !c->visible)
			return (0);
	}
	return (1);
}

static int
dostock(void)
{
	struct card *c;

	if (stock.last == NULL) {
		/* Copy entire stock back. */
		while ((c = waste.last) != NULL) {
			cs_move_tail(&waste, &stock, c);
			c->visible = 0;
		}
	}

	c = stock.last;
	if (c == NULL)
		return (1);

	/* Prevent us from going back too far. */
	if (waste.last != NULL && waste.last->prev != NULL &&
	    waste.last->prev->prev != NULL)
		waste.last->prev->prev->visible = 0;

	/* Pick a single card. */
	cs_move_tail(&stock, &waste, c);
	c->visible = 1;
	return (0);
}

static int
domove(struct cardstack *fs, struct cardstack *ts)
{
	struct card *fl = fs->last, *tl = ts->last;

	if (fs == ts)
		return (1);

	if (ts->type == CST_FOUNDATION) {
		/* To the foundations. */
		if (!okay_foundation(fl, tl))
			return (1);
	} else if (ts->type == CST_TABLEAU) {
		if (fs->type == CST_TABLEAU) {
			/* From tableaus to tableaus. */
			while (!okay_tableau(fl, tl)) {
				fl = fl->prev;
				if (fl == NULL || !fl->visible)
					return (1);
			}
		} else {
			/* From other stacks to tableaus. */
			if (!okay_tableau(fl, tl))
				return (1);
		}
	} else {
		return (1);
	}

	cs_move_tail(fs, ts, fl);

	/* Removal from tableaus causes next card to become visible. */
	fl = fs->last;
	if (fs->type == CST_TABLEAU && fl != NULL)
		fl->visible = 1;
	return (0);
}

static void
dostep(void)
{
	struct cardstack *cs;
	struct card *c;
	int ch;

	ch = getch();
	if ((ch == 'x') || (ch == 27)) {
	  endterm();
	  exit (0);
	}
	if (ch == 'q') {
		if (dostock() != 0)
			beep();
	} else {
		cs = getstack(ch);
		if (cs == NULL) {
			beep();
			stepfrom = NULL;
		} else if (stepfrom == NULL) {
			c = cs->last;
			if (c == NULL || !c->visible)
				beep();
			else
				stepfrom = cs;
		} else {
			if (domove(stepfrom, cs) != 0)
				beep();
			stepfrom = NULL;
		}
	}
}

int
main(int argc __unused, char *argv[] __unused)
{

	initstacks();
	initcards();
	loadstacks();

	initterm();
	while (!finished()) {
		display();
		dostep();
	}
	endterm();
	return (0);
}
